@@ -22,16 +22,35 @@
 
 
 
-;;; the following useless constants are just ridiculously high because there is no actual limit
-(defconstant LAMBDA-PARAMETERS-LIMIT 4294967296)
-(defconstant CALL-ARGUMENTS-LIMIT 4294967296)
-(defconstant MULTIPLE-VALUES-LIMIT 4294967296)
-(defconstant ARRAY-RANK-LIMIT 4294967296)
-(defconstant ARRAY-DIMENSION-LIMIT 4294967296)
-(defconstant ARRAY-TOTAL-SIZE-LIMIT 4294967296)
+(defparameter *features* '(:common-lisp))
 
 
-(defconstant CHAR-CODE-LIMIT 4294967296)
+
+(defun machine-instance nil)
+(defun machine-type nil)
+(defun machine-version nil)
+
+
+;;; the following useless constants are just ridiculously high because lambda
+;;; lists and multiple values use chained lists, so there's no hard limit
+
+(defconstant LAMBDA-PARAMETERS-LIMIT most-positive-fixnum)
+(defconstant CALL-ARGUMENTS-LIMIT most-positive-fixnum)
+(defconstant MULTIPLE-VALUES-LIMIT most-positive-fixnum)
+
+
+
+(defconstant ARRAY-RANK-LIMIT most-positive-fixnum)
+(defconstant ARRAY-DIMENSION-LIMIT most-positive-fixnum)
+(defconstant ARRAY-TOTAL-SIZE-LIMIT most-positive-fixnum)
+
+
+(defconstant CHAR-CODE-LIMIT most-positive-fixnum)
+
+
+
+(defconstant lambda-list-keywords '(&optional &rest &body &key &allow-other-keys))
+
 
 
 (defun identity (x) x)
@@ -188,10 +207,33 @@
 	     (cond ,@rest)))))
 
 
+(defmacro case (keyf &rest clauses)
+  (let ((keysym (gensym))
+	outsym)
+    `(let ((,keysym ,keyf))
+       ,(dolist (clause clauses (setq outsym (cons 'cond outsym)))
+	  (cond ((or (eq (car clause) t)
+		     (eq (car clause) 'otherwise))
+		 (setq outsym (append outsym `((t (progn ,@(cdr clause)))))))
+		((listp (car clause))
+		 (setq outsym (append outsym `(((member ,keysym ',(car clause))
+						(progn ,@(cdr clause)))))))
+		(t
+		 (setq outsym (append outsym `(((eql ,keysym ',(car clause))
+						(progn ,@(cdr clause))))))))))))
+
+
 (defmacro return (&optional val)
   `(return-from nil (values-list (multiple-value-list ,val))))
 
 
+(defmacro multiple-value-bind (vars valform &body forms)
+  (let ((restsym (gensym)))
+    `(multiple-value-call (lambda (&optional ,@vars &rest ,restsym)
+			    ,@forms)
+       ,valform)))
+
+
 (defun every (pred &rest sequences)
   (let ((n (apply 'min (mapcar #'length sequences))))
     (dotimes (i n)
@@ -447,6 +489,47 @@
   str)
 
 
+(defun string-left-trim (charbag s)
+  (let ((str (string s)))
+    (dotimes (i (length str) (subseq str 0 0))
+      (if (not (position (aref str i) charbag))
+	  (return-from string-left-trim (subseq str i))))))
+
+
+(defun string-right-trim (charbag s)
+  (let* ((str (string s))
+	 (len (length str)))
+    (dotimes (i len (subseq str 0 0))
+      (if (not (position (aref str (- len i 1)) charbag))
+	  (return-from string-right-trim (subseq str 0 (- len i)))))))
+
+
+(defun string-trim (charbag s)
+  (string-left-trim charbag (string-right-trim charbag s)))
+
+
+
+(defmacro defpackage (s &rest opts)
+  (let ((name (string s))
+	nicks uses imports exports intern)
+    (dolist (opt opts)
+      (case (car opt)
+	(:nicknames (setq nicks (append nicks (mapcar #'string (cdr opt)))))
+	(:use (setq uses (append uses (mapcar #'find-package (cdr opt)))))
+	(:import-from (setq imports (append imports (list (list (find-package (cadr opt)) (mapcar #'string (cddr opt)))))))
+	(:export (setq exports (append exports (mapcar #'string (cdr opt)))))
+	(:intern (setq intern (append intern (mapcar #'string (cdr opt)))))))
+    `(progn
+       (unless (find-package ,name)
+	 (make-package ,name))
+       (rename-package ,name ,name ',nicks)
+       (use-package ',uses ,name)
+       ,@(mapcar (lambda (l) `(import (list ,@(mapcar (lambda (sm) `(intern ,sm ,(car l))) (cadr l))) ,name)) imports)
+       ,@(mapcar (lambda (sm) `(intern ,sm ,name)) intern)
+       (export (list ,@(mapcar (lambda (st) `(intern ,st ,name)) exports)) ,name)
+       (find-package ,name))))
+
+
 
 (defun consp (obj)
   (typep obj 'cons))
@@ -513,6 +596,16 @@
 
 
 
+(defun macroexpand (form)
+  (do ((ret (list form t))
+       expp)
+      ((not (cadr ret)) (values (car ret) expp))
+    (setq ret (multiple-value-list (macroexpand-1 (car ret))))
+    (if (and (not expp) (cadr ret))
+	(setq expp t))))
+
+
+
 (defun equal (x y)
   (cond
     ((numberp x) (eql x y))
@@ -608,6 +701,97 @@
   obj)
 
 
+(defmacro loop (&body forms)
+  (let (block-name
+	(tagname (gensym))
+	(inner-block-name (gensym))
+	initially-forms
+	finally-forms
+	do-forms
+	vars var)
+    (do ((forms forms (cdr forms)))
+	((not forms))
+      (let ((form (car forms)))
+	(if (atom form)
+	    (let ((sym (string form)))
+	      (cond
+                 ((string= sym "INITIALLY")
+		 (do ((f (cdr forms)))
+		     ((atom (car f)))
+		   (setq initially-forms (append initially-forms `(,(car f))))
+		   (setq f (cdr f))))
+		((string= sym "FINALLY")
+		 (do ((f (cdr forms)))
+		     ((atom (car f)))
+		   (setq finally-forms (append finally-forms `(,(car f))))
+		   (setq f (cdr f))))
+		((or (string= sym "DO")
+		     (string= sym "DOING"))
+		 (do ((f (cdr forms)))
+		     ((atom (car f)))
+		   (setq do-forms (append do-forms `(,(car f))))
+		   (setq f (cdr f))))
+		((string= sym "NAMED")
+		 (setq block-name (cadr forms))
+		 (setq forms (cdr forms)))
+		((or (string= sym "FOR")
+		     (string= sym "AS"))
+		 (setq forms (cdr forms))
+		 (setq var `(,(car forms) 0 nil 1 1))
+		 (setq vars (append vars `(,var))))
+		((string= sym "FROM")
+		 (setq forms (cdr forms))
+		 (setf (elt var 1) (car forms)))
+		((or (string= sym "DOWNFROM"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 1) (car forms))
+		 (setf (elt var 3) -1))
+		((or (string= sym "UPFROM"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 1) (car forms))
+		 (setf (elt var 3) 1))
+		((or (string= sym "TO"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 2) (car forms)))
+		((or (string= sym "DOWNTO"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 2) (car forms))
+		 (setf (elt var 3) -1))
+		((or (string= sym "UPTO"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 2) (car forms))
+		 (setf (elt var 3) 1))
+		((or (string= sym "BELOW"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 2) (1- (car forms))))
+		((or (string= sym "ABOVE"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 2) (1+ (car forms))))
+		((or (string= sym "BY"))
+		 (setq forms (cdr forms))
+		 (setf (elt var 4) (car forms)))))
+	    (append do-forms `(,form)))))
+    `(block ,block-name
+       (tagbody
+	  (let ,(mapcar (lambda (x) `(,(car x) ,(cadr x))) vars)
+	    ,@initially-forms
+	    (block ,inner-block-name
+	      (tagbody
+		 ,tagname
+		 (if (or ,@(mapcar (lambda (x) (if (elt x 2)
+						      (if (= (elt x 3) 1)
+							  `(> ,(elt x 0) ,(elt x 2))
+							  `(< ,(elt x 0) ,(elt x 2)))
+						      nil)) vars))
+		     (return-from ,inner-block-name nil))
+		 ,@do-forms
+		 (progn
+		   ,@(mapcar (lambda (x) `(setq ,(elt x 0) (+ ,(elt x 0) ,(* (elt x 3) (elt x 4))))) vars))
+		 (go ,tagname)))
+	    ,@finally-forms
+	    nil)))))
+
+
 (defun format (out fstr &rest args)
   (let (in-spec at-sign colon sign num dirargs)
     (dotimes (i (length fstr))
@@ -629,14 +813,16 @@
 
 
 
-(export '(lambda-parameters-limit call-arguments-limit multiple-values-limit
+(export '(*features* machine-instance machine-type machine-version
+	  lambda-parameters-limit call-arguments-limit multiple-values-limit
 	  array-rank-limit array-dimension-limit array-total-size-limit
-	  char-code-limit identity pi 1+ 1- minusp plusp abs zerop mod rem evenp
-	  oddp isqrt conjugate *gensym-counter* gensym first second third fourth
-	  fifth sixth seventh eighth ninth tenth rest caar cadr cdar cddr caaar
-	  caadr cadar caddr cdaar cdadr cddar cdddr caaaar caaadr caadar caaddr
-	  cadaar cadadr caddar cadddr cdaaar cdaadr cdadar cdaddr cddaar cddadr
-	  cdddar cddddr endp when unless incf decf and or cond return every some
+	  char-code-limit lambda-list-keywords identity pi 1+ 1- minusp plusp
+	  abs zerop mod rem evenp oddp isqrt conjugate *gensym-counter* gensym
+	  first second third fourth fifth sixth seventh eighth ninth tenth rest
+	  caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
+	  caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
+	  cdadar cdaddr cddaar cddadr cdddar cddddr endp when unless incf decf
+	  and or cond otherwise case return multiple-value-bind every some
 	  notany notevery member member-if member-if-not find find-if
 	  find-if-not assoc assoc-if assoc-if-not position position-if
 	  position-if-not count count-if count-if-not remove remove-if-not
@@ -644,11 +830,12 @@
 	  array-rank array-dimension array-total-size array-in-bounds-p char
 	  string/= char-equal digit-char digit-char-p char-int string-upcase
 	  string-downcase string-capitalize nstring-upcase nstring-downcase
-	  nstring-capitalize consp listp symbolp keywordp functionp packagep
-	  integerp rationalp floatp complexp characterp vectorp arrayp sequencep
-	  stringp hash-table-p pathnamep streamp realp numberp equal equalp
+	  nstring-capitalize string-left-trim string-right-trim string-trim
+	  defpackage consp listp symbolp keywordp functionp packagep integerp
+	  rationalp floatp complexp characterp vectorp arrayp sequencep stringp
+	  hash-table-p pathnamep streamp realp numberp macroexpand equal equalp
 	  fdefinition complement mapc terpri write-line write-sequence prin1
-	  princ print format))
+	  princ print loop format))
 
 
 
