@@ -185,6 +185,8 @@ make_test ("(values (+ 1 0) 2 3)", "1\n2\n3");
 make_test ("(values-list '(1 2 3))", "1\n2\n3");
 make_test ("(multiple-value-list (values))", "NIL");
 make_test ("(multiple-value-list (values 1 2 3))", "(1 2 3)");
+make_test ("(multiple-value-call (lambda ()))", "NIL");
+make_test ("(multiple-value-call #'+ (floor 5 2) (values) 1)", "4");
 make_test ("(if nil 1)", "NIL");
 make_test ("(common-lisp:if '(1) 2)", "2");
 make_test ("(cl:eq 1 1)", "NIL");
@@ -205,6 +207,7 @@ make_test ("(list 1 2 3)", "(1 2 3)");
 make_test ("(list* 0)", "0");
 make_test ("(list* 0 1 2)", "(0 1 . 2)");
 make_test ("(append ())", "NIL");
+make_test ("(append nil nil 0)", "0");
 make_test ("(append '(1 2) '(3 4))", "(1 2 3 4)");
 make_test ("(append '(1) (values))", "(1)");
 make_test ("(nconc)", "NIL");
@@ -215,11 +218,27 @@ make_test ("(nconc (list 'a 'b 'c) NIL (list 'd 'e) NIL 'f)", "(A B C D E . F)")
 make_test ("(let ((a 2)) a)", "2");
 make_test ("(let ((x 0)) x x x)", "0");
 make_test ("(let ((a)) a)", "NIL");
+make_test ("(let () (declare (ignore a b (function c)) (ignorable x (function y))) (declare (ignore)) (+ 1 2))", "3");
+make_test ("(let () (declare (inline a b (setf c)) (notinline x (setf y))) (+ 1 2))", "3");
+make_test ("(let () (declare (optimize safety speed (compilation-speed 0) (debug 3))) (+ 1 2))", "3");
+make_test ("(let ((x 0)) (declare (special x)) x)", "0");
+make_test ("(let ((x 0)) (declare (special x)) (let ((x 1)) x))", "1");
+make_test ("(let ((x 0)) (declare (special x)) (let ((x 1)) (eval 'x)))", "0");
+make_test ("(let ((x 0)) (let ((x 1)) (declare (special x)) (let () x)))", "1");
+make_test ("(defun func () x)", "FUNC");
+make_test ("(let ((x 10)) (declare (special x)) (func))", "10");
+make_test ("(let ((x 12)) (declare (special x)) (let ((x 13)) (write (func)) (write x)))", "1213\n13");
+make_test ("(let* ((x 0)) (declare (special x)) x)", "0");
+make_test ("(let* ((x 0)) (declare (special x)) (let ((x 1)) x))", "1");
+make_test ("(let* ((x 0)) (declare (special x)) (let ((x 1)) (eval 'x)))", "0");
 make_test ("(let* ((a)) a)", "NIL");
 make_test ("(flet ((a () (write \"\"))) (a))", "\"\"\n\"\"");
 make_test ("(labels ((a NIL (write \"\"))) (a))", "\"\"\n\"\"");
 make_test ("(labels ((a () (b)) (b () (write \"\"))) (a))", "\"\"\n\"\"");
 make_test ("(macrolet ((w (a) `(write ,a))) (w \"\"))", "\"\"\n\"\"");
+make_test ("(eval-when (:compile-toplevel compile) 0)", "NIL");
+make_test ("(eval-when (:load-toplevel load) 0)", "NIL");
+make_test ("(eval-when (:execute eval load compile) 0)", "0");
 make_test ("(defconstant a 8)", "A");
 make_test ("(defconstant a 8)", "A");
 make_test ("a", "8");
@@ -268,6 +287,7 @@ make_test ("(apply #'+ '(1 2))", "3");
 make_test ("(apply #'car '(1 2) ())", "1");
 make_test ("(apply 'car '(1 2) ())", "1");
 make_test ("(apply (lambda (x) (car x)) '(1 2) ())", "1");
+make_test ("(let ((s 0)) (funcall (lambda (s) s) 1))", "1");
 make_test ("(funcall #'+ 1 2 3)", "6");
 make_test ("(funcall 'car '(1 2 3))", "1");
 make_test ("(f)", "NIL");
@@ -298,21 +318,27 @@ make_test ("(defun fun (&rest x) x)", "FUN");
 make_test ("(fun)", "NIL");
 make_test ("(defun fun (x &rest y) y)", "FUN");
 make_test ("(fun 5 6 7 8)", "(6 7 8)");
-make_test ("(defun fun2 (&rest r &key a b c) b)", "FUN2");
-make_test ("(fun2 :b 0)", "0");
-make_test ("(fun2 :a 0)", "NIL");
-make_test ("(fun2 :c 'tt :b (+ 1 2))", "3");
+make_test ("(defun fun2 (&rest r &key x y z) y)", "FUN2");
+make_test ("(fun2 :y 0)", "0");
+make_test ("(fun2 :y 0 :y 1)", "0");
+make_test ("(fun2 :x 0)", "NIL");
+make_test ("(fun2 :z 'tt :y (+ 1 2))", "3");
 make_test ("(fun2)", "NIL");
 make_test ("(defun fun2 (&key ((:k aa) 1 supp)) (list aa supp))", "FUN2");
 make_test ("(fun2)", "(1 NIL)");
 make_test ("(fun2 :k 1)", "(1 T)");
-make_test ("(defun fun2 (&rest r &key a b c) r)", "FUN2");
-make_test ("(fun2 :b (+ 1 2) :c 'tt)", "(:B 3 :C TT)");
+make_test ("(defun fun2 (&rest r &key x y z) r)", "FUN2");
+make_test ("(fun2 :y (+ 1 2) :z 'tt)", "(:Y 3 :Z TT)");
 make_test ("(fun2)", "NIL");
 make_test ("(defun fun (x &rest y &key z) y)", "FUN");
 make_test ("(fun 1 :z 0)", "(:Z 0)");
 make_test ("(defun fun (x &optional y &rest z &key k) z)", "FUN");
 make_test ("(fun 0 1 :k 2)", "(:K 2)");
+make_test ("(defun f (&key &allow-other-keys))", "F");
+make_test ("(f :a 0)", "NIL");
+make_test ("(defun f (&key x &allow-other-keys) x)", "F");
+make_test ("(f :x 0 :y 1)", "0");
+make_test ("(f)", "NIL");
 make_test ("(let ((x 0)) (defun inc () (setf x (+ x 1))))", "INC");
 make_test ("(inc)", "1");
 make_test ("(inc)", "2");
@@ -339,6 +365,15 @@ make_test ("(1- .5)", "-0.5");
 make_test ("(byte 20 100)", "#<BYTE-SPECIFIER size 20 position 100>");
 make_test ("(byte-size (byte 20 100))", "20");
 make_test ("(byte-position (byte 20 100))", "100");
+
+make_test ("(macroexpand-1 '(incf var))", "(SETF VAR (+ VAR 1))\nT");
+make_test ("(macroexpand-1 '(+ 0))", "(+ 0)\nNIL");
+make_test ("(defmacro test2 nil (+ 1 2))", "TEST2");
+make_test ("(defmacro test nil '(test2))", "TEST");
+make_test ("(macroexpand '(test2))", "3\nT");
+make_test ("(macroexpand '(test))", "3\nT");
+make_test ("(macroexpand '(+ 1 2))", "(+ 1 2)\nNIL");
+
 make_test ("(typep '(1 . 2) 'cons)", "T");
 make_test ("(typep () 'atom)", "T");
 make_test ("(typep (cons 1 2) 'atom)", "NIL");
@@ -377,8 +412,35 @@ make_test ("(typep #p\"\" 'pathname)", "T");
 make_test ("(typep (open #p\"README\") 'stream)", "T");
 make_test ("(typep (open #p\"README\") 'file-stream)", "T");
 make_test ("(typep (make-string-input-stream \"hello\") 'string-stream)", "T");
+make_test ("(type-of 'a)", "SYMBOL");
+make_test ("(type-of \"aaa\")", "STRING");
+make_test ("(type-of #\\a)", "CHARACTER");
+make_test ("(type-of '(1 . 2))", "CONS");
+make_test ("(type-of 0)", "INTEGER");
+make_test ("(type-of 1/2)", "RATIO");
+make_test ("(type-of 1.0)", "FLOAT");
+make_test ("(type-of (complex 0 1))", "COMPLEX");
+make_test ("(type-of #'car)", "FUNCTION");
+make_test ("(type-of *package*)", "PACKAGE");
+make_test ("(type-of #(0 1 2))", "ARRAY");
+make_test ("(type-of (make-hash-table))", "HASH-TABLE");
+make_test ("(type-of #p\"aaa\")", "PATHNAME");
+make_test ("(type-of *standard-output*)", "STREAM");
 make_test ("(deftype not-integer () `(not integer))", "NOT-INTEGER");
+
 make_test ("(make-string 3)", "\"\0\0\0\"");
+make_test ("(intern \"hi\")", "|hi|\nNIL");
+make_test ("(intern \"hi\")", "|hi|\n:INTERNAL");
+make_test ("(intern \"hi\" 'keyword)", ":|hi|\nNIL");
+make_test ("(intern \"hi\" 'keyword)", ":|hi|\n:EXTERNAL");
+make_test ("(find-symbol \"CAR\")", "CAR\n:INHERITED");
+make_test ("(find-symbol \"CAR\" 'cl)", "CAR\n:EXTERNAL");
+make_test ("(find-symbol \"car\")", "NIL\nNIL");
+make_test ("(unintern :hi)", "NIL");
+make_test ("(unintern '|hi|)", "T");
+make_test ("(intern \"hi\")", "|hi|\nNIL");
+make_test ("(unintern :|hi| 'keyword)", "T");
+make_test ("(intern \"hi\" 'keyword)", ":|hi|\nNIL");
 make_test ("(make-symbol \"aaa\")", "#:|aaa|");
 make_test ("(boundp 'b)", "T");
 make_test ("(boundp 'g3)", "NIL");
@@ -456,6 +518,11 @@ make_test ("(digit-char-p #\\g 16)", "NIL");
 make_test ("(string-upcase \"Hello\")", "\"HELLO\"");
 make_test ("(string-downcase \"Hello\")", "\"hello\"");
 make_test ("(string-capitalize \"this is a Good day\")", "\"This Is A Good Day\"");
+make_test ("(string-left-trim '(#\\a #\\b) \"aabbabahello\")", "\"hello\"");
+make_test ("(string-left-trim '(#\\a #\\b) \"aabbab\")", "\"\"");
+make_test ("(string-right-trim '(#\\a #\\b) \"helloaabbab\")", "\"hello\"");
+make_test ("(string-right-trim '(#\\a #\\b) \"aabbab\")", "\"\"");
+make_test ("(string-trim '(#\\a #\\b) \"baahelloaabbab\")", "\"hello\"");
 make_test ("(car '(1 2))", "1");
 make_test ("(cdr '(1 2))", "(2)");
 make_test ("(car ())", "NIL");
@@ -484,6 +551,14 @@ make_test ("(aref (make-array '(1 2 3)) 0 1 2)", "NIL");
 make_test ("(row-major-aref #(0 1 2) 1)", "1");
 make_test ("(row-major-aref (make-array '(3 2 1)) 5)", "NIL");
 make_test ("(char \"àbcdef\" 2)", "#\\c");
+make_test ("(subseq nil 0)", "NIL");
+make_test ("(subseq \"hello\" 1)", "\"ello\"");
+make_test ("(subseq \"hello\" 1 3)", "\"el\"");
+make_test ("(subseq \"hello\" 1 1)", "\"\"");
+make_test ("(subseq #(0 1 2) 1 3)", "#(1 2)");
+make_test ("(subseq #(0 1 2) 1 1)", "#()");
+make_test ("(subseq '(0 1 2) 1 nil)", "(1 2)");
+make_test ("(subseq '(0 1 2) 1 1)", "NIL");
 make_test ("(length ())", "0");
 make_test ("(length \"aaa\")", "3");
 make_test ("(length #(1 2))", "2");
@@ -585,15 +660,23 @@ make_test ("(princ '|Aa,a|)", "Aa,a\n|Aa,a|");
 make_test ("(princ #\\a)", "a\n#\\a");
 make_test ("(princ \"aaa\" *standard-output*)", "aaa\n\"aaa\"");
 make_test ("(print \"aaa\" *standard-output*)", "\n\"aaa\" \n\"aaa\"");
+make_test ("(loop initially (write 'hello) for i from 1 to 5 finally (write 'world) do (write i))", "HELLO12345WORLD\nNIL");
+make_test ("(loop for i from 1 to 5 do (write i))", "12345\nNIL");
+make_test ("(loop for i from 1 below 5 do (write i))", "1234\nNIL");
+make_test ("(loop for i from 10 downto 5 by 2 do (write i))", "1086\nNIL");
+make_test ("(loop for i upfrom 5 to 10 do (write i))", "5678910\nNIL");
+make_test ("(loop for i from 5 upto 11 for j downfrom 20 above 15 do (write i) (write j))", "520619718817916\nNIL");
+make_test ("(loop named foo for i from 0 do (write i) (if (> i 4) (return-from foo 0)))", "012345\n0");
+make_test ("(loop as i from 0 do (if (> i 5) (return 10) (write i)) finally (write 'finished))", "012345\n10");
 make_test ("(format t \"aaa\")", "aaa\nNIL");
 make_test ("(format t \"aaa~~\")", "aaa~\nNIL");
 make_test ("(format t \"aaa~%~%\")", "aaa\n\nNIL");
 make_test ("(format t \"aa~&~&bb\")", "aa\nbb\nNIL");
 make_test ("(format t \"the number is ~a and the list is ~s\" 10 '(1 2 3))",
 	   "the number is 10 and the list is (1 2 3)\nNIL");
-make_test ("(defparameter str (open \"writetest\" :direction :output))", "STR");
+make_test ("(defparameter str (open \"writetest\" :direction :output :direction :input))", "STR");
 make_test ("(write \"hello\" :stream str)", "\"hello\"");
-make_test ("(write #\\a :stream str)", "#\\a");
+make_test ("(write #\\a :stream str :stream *standard-input*)", "#\\a");
 make_test ("(write 1 :stream str)", "1");
 make_test ("(write 1/2 :stream str)", "1/2");
 make_test ("(write 0.1 :stream str)", "0.1");
@@ -612,11 +695,11 @@ make_test ("(write 100 :stream str)", "100");
 make_test ("(write 1.2 :stream str)", "1.2");
 make_test ("(get-output-stream-string str)", "\"AAA\\\"bbb\\\"CCC#\\\\c1001.2\"");
 make_test ("(get-output-stream-string str)", "\"\"");
-make_test ("(gensym)", "#:G1");
-make_test ("(gensym)", "#:G2");
+make_test ("(gensym)", "#:G17");
+make_test ("(gensym)", "#:G18");
 make_test ("(gensym 5)", "#:G5");
-make_test ("(gensym \"A\")", "#:A3");
-make_test ("(gensym)", "#:G4");
+make_test ("(gensym \"A\")", "#:A19");
+make_test ("(gensym)", "#:G20");
 make_test ("(second '(0))", "NIL");
 make_test ("(fifth '(0 1 2 3 4))", "4");
 make_test ("(endp '(1 . 2))", "NIL");
@@ -687,6 +770,7 @@ make_test ("(dolist (i '(0 1 2 3)) (if (= i 2) (return (values 12 13))) (write i
 make_test ("(mapcar #'car '((1 a) (2 b) (3 c)))", "(1 2 3)");
 make_test ("(mapcar #'abs '(3 -4 2 -5 -6))", "(3 4 2 5 6)");
 make_test ("(mapcar #'cons '(a b c) '(1 2 3))", "((A . 1) (B . 2) (C . 3))");
+make_test ("(let ((s 0)) (write (mapcar (lambda (s) s) '(1 2 3))) s)", "(1 2 3)\n0");
 make_test ("(remove 3 '(1 2 3 4 3 2 1))", "(1 2 4 2 1)");
 make_test ("(remove #\\e \"abcdefg\")", "\"abcdfg\"");
 make_test ("(remove-if #'numberp #(1 2 3))", "#()");
@@ -708,6 +792,15 @@ make_test ("(pop ll)", "1");
 make_test ("ll", "NIL");
 make_test ("(cond (t (+ 1 2)))", "3");
 make_test ("(cond (nil 2) (t 3) (nil 4))", "3");
+make_test ("(cond ((= 2 (+ 1 2)) (+ 0 10)) ((= 3 (+ 1 2)) (+ 12) (+ 0 11)))", "11");
+make_test ("(case (+ 1 2) (3 (+ 1 5)))", "6");
+make_test ("(case (+ 1 2) (4 (+ 1 5)))", "NIL");
+make_test ("(case (+ 1 2) (2 10) ((4 3) (+ 1 5)))", "6");
+make_test ("(case (+ 1 2) (2 10) ((4 3) (+ 1 5)) (otherwise 12))", "6");
+make_test ("(case (+ 1 2) (2 10) ((4 5) (+ 1 5)) (otherwise 11 12))", "12");
+make_test ("(multiple-value-bind (x y) (floor 2 1) (list x y))", "(2 0)");
+make_test ("(multiple-value-bind (x) (floor 2 1) (list x))", "(2)");
+make_test ("(multiple-value-bind (x y z) (floor 2 1) (list x y z))", "(2 0 NIL)");
 make_test ("(every #'evenp '(0 2 4))", "T");
 make_test ("(every #'< '(0 1 2) '(1 1))", "NIL");
 make_test ("(some #'< '(0 1 2) '(1 1))", "T");
@@ -735,6 +828,31 @@ make_test ("(count #\\a \"abcabc\")", "2");
 make_test ("(count 2 '(1 2 3))", "1");
 make_test ("(count-if #'upper-case-p \"aAbBcCdD\")", "4");
 make_test ("(count-if-not #'evenp #(0 1 2))", "1");
+make_test ("(machine-type)", "NIL");
+make_test ("(machine-instance)", "NIL");
+make_test ("(machine-version)", "NIL");
+
+make_test ("#+common-lisp \"\" \"\"", "\"\"\n\"\"");
+make_test ("#-common-lisp \"\" \"\"", "\"\"");
+make_test ("#+(and common-lisp foo) \"\" \"\"", "\"\"");
+make_test ("#+(or common-lisp foo) \"\" \"\"", "\"\"\n\"\"");
+make_test ("#-(not common-lisp) \"\" \"\"", "\"\"\n\"\"");
+make_test ("#+common-lisp 'x", "X");
+make_test ("#+common-lisp 0", "0");
+make_test ("#-common-lisp 0 0", "0");
+make_test ("#-common-lisp 'aaa 'bbb", "BBB");
+make_test ("#+common-lisp '(1 2 3) \"\"", "(1 2 3)\n\"\"");
+make_test ("#+foo '(1 2 3) \"\"", "\"\"");
+make_test ("#+foo '((1 2) 3) 0", "0");
+make_test ("#+(and\ncommon-lisp) 0 0", "0\n0");
+make_test ("#-(\nand common-lisp) 0 0", "0");
+make_test ("#+\n(and common-lisp) 0 0", "0\n0");
+make_test ("#+common-lisp (\n)", "NIL");
+make_test ("#+common-lisp '(1\n2)", "(1 2)");
+make_test ("#-common-lisp (\n) 0", "0");
+make_test ("#+foo \"aaa\nbbb\" 0", "0");
+make_test ("#+foo (1\n2 (3)) 0", "0");
+
 make_test ("*package*", "#<PACKAGE \"COMMON-LISP-USER\">");
 make_test ("(find-package *package*)", "#<PACKAGE \"COMMON-LISP-USER\">");
 make_test ("(find-package \"CL\")", "#<PACKAGE \"COMMON-LISP\">");
@@ -751,6 +869,7 @@ make_test ("(package-used-by-list (symbol-package 'if))", "(#<PACKAGE \"COMMON-L
 make_test ("(list-all-packages)", "(#<PACKAGE \"COMMON-LISP-USER\"> #<PACKAGE \"COMMON-LISP\"> #<PACKAGE \"KEYWORD\">)");
 make_test ("(make-package \"test\")", "#<PACKAGE \"test\">");
 make_test ("(list-all-packages)", "(#<PACKAGE \"test\"> #<PACKAGE \"COMMON-LISP-USER\"> #<PACKAGE \"COMMON-LISP\"> #<PACKAGE \"KEYWORD\">)");
+make_test ("(rename-package \"test\" \"test\")", "#<PACKAGE \"test\">");
 make_test ("(rename-package \"test\" \"TEST\" '(nick1 #\\n \"nick3\"))", "#<PACKAGE \"TEST\">");
 make_test ("(list-all-packages)", "(#<PACKAGE \"TEST\"> #<PACKAGE \"COMMON-LISP-USER\"> #<PACKAGE \"COMMON-LISP\"> #<PACKAGE \"KEYWORD\">)");
 make_test ("(package-nicknames \"TEST\")", "(\"NICK1\" \"n\" \"nick3\")");
@@ -763,19 +882,28 @@ make_test ("(make-package \"newtest\")", "#<PACKAGE \"newtest\">");
 make_test ("(in-package \"newtest\")", "#<PACKAGE \"newtest\">");
 make_test ("(cl:package-use-list cl:*package*)", "COMMON-LISP:NIL");
 make_test ("(cl:use-package 'cl)", "T");
+make_test ("(cl:use-package nil \"newtest\")", "T");
 make_test ("(cl:package-use-list cl:*package*)", "(#<PACKAGE \"COMMON-LISP\">)");
 make_test ("(cadr '(0 1))", "1");
 make_test ("(cl:import 'cl:car)", "T");
 make_test ("(cl:import '(cl:car) '|newtest|)", "T");
+make_test ("(cl:export nil \"newtest\")", "T");
 make_test ("(cl:export 'jjj)", "T");
 make_test ("(cl:export '(jjj) cl:*package*)", "T");
 make_test ("'|newtest|:jjj", "JJJ");
+make_test ("(cl:unexport nil \"newtest\")", "T");
 make_test ("(cl:unexport '(jjj) cl:*package*)", "T");
 make_test ("(car ())", "NIL");
 make_test ("(cl:unuse-package 'cl)", "COMMON-LISP:T");
+make_test ("(cl:unuse-package cl:nil \"newtest\")", "COMMON-LISP:T");
 make_test ("(cl:package-use-list cl:*package*)", "COMMON-LISP:NIL");
 make_test ("(car ())", "COMMON-LISP:NIL");
 make_test ("(cl:in-package cl-user)", "#<PACKAGE \"COMMON-LISP-USER\">");
+make_test ("(defpackage \"aaa\" (:nicknames \"bbb\" |ccc|) (:use cl-user) (:export \"sym\") (:intern \"eee\" \"fff\") (:import-from \"CL\" car caar) (:import-from \"CL-USER\" al-argc))", "#<PACKAGE \"aaa\">");
+make_test ("(package-nicknames \"aaa\")", "(\"bbb\" \"ccc\")");
+make_test ("(package-use-list \"aaa\")", "(#<PACKAGE \"COMMON-LISP-USER\">)");
+make_test ("'|aaa|:|sym|", "|aaa|:|sym|");
+make_test ("(defpackage \"aaa\" (:nicknames \"bbb\" |ccc|) (:use cl-user) (:export \"sym\") (:intern \"eee\" \"fff\") (:import-from \"CL\" car caar) (:import-from \"CL-USER\" al-argc))", "#<PACKAGE \"aaa\">");
 
 
 # arithmetic tests
@@ -835,6 +963,8 @@ make_test ("(complex 1 0)", "1");
 make_test ("(complex 1 0.0)", "#C(1.0 0.0)");
 make_test ("(complex 1.0 1)", "#C(1.0 1.0)");
 make_test ("(complex 1 1.0)", "#C(1.0 1.0)");
+make_test ("#c (0 1)", "#C(0 1)");
+make_test ("#C(1 1.0)", "#C(1.0 1.0)");
 make_test ("(realpart 1)", "1");
 make_test ("(realpart (complex 1 2.0))", "1.0");
 make_test ("(imagpart 1)", "0");
@@ -939,7 +1069,7 @@ sub make_test
 
 	if (not defined ($out) and eof ($al_out))
 	{
-	    print "\n\nEOF reached from al, it probably crashed\n";
+	    print "\n\nGot EOF from al, it probably crashed\n";
 
 	    exit;
 	}
